/* FILE: threevector.h        -*-Mode: c++-*-
 *
 * Simple 3D vector class.  *All* client code should use the
 * ThreeVector typedef in case this class gets moved into an
 * OOMMF extension library.
 *
 */

#ifndef _OXS_THREEVECTOR
#define _OXS_THREEVECTOR

#include "oc.h"

OC_USE_STD_NAMESPACE;  // For std::fabs()

/* End includes */

class Oxs_ThreeVector;  // Forward declaration for typedef
typedef Oxs_ThreeVector ThreeVector;

class Oxs_ThreeVector {
public:
  REAL8m x,y,z;

  // Constructors
  Oxs_ThreeVector(): x(0.), y(0.), z(0.) {}
  Oxs_ThreeVector(REAL8m xi,REAL8m yi,REAL8m zi)
    : x(xi), y(yi), z(zi) {}
  Oxs_ThreeVector(REAL8m *arr) : x(arr[0]), y(arr[1]), z(arr[2]) {}
  Oxs_ThreeVector(const Oxs_ThreeVector &v) : x(v.x), y(v.y), z(v.z) {}

  // Assignment operators
  Oxs_ThreeVector& Set(REAL8m xi,REAL8m yi,REAL8m zi)
    { x=xi; y=yi; z=zi; return *this; }
  Oxs_ThreeVector& operator=(const Oxs_ThreeVector &v) {
    x=v.x; y=v.y; z=v.z;
    return *this;
  }
  Oxs_ThreeVector& operator+=(const Oxs_ThreeVector& v) { 
    x+=v.x; y+=v.y; z+=v.z;
    return *this;
  }
  Oxs_ThreeVector& operator-=(const Oxs_ThreeVector& v) {
    x-=v.x; y-=v.y; z-=v.z;
    return *this;
  }

  Oxs_ThreeVector& operator^=(const Oxs_ThreeVector& w) { // Cross product
#if 0
    REAL8m tx = y * w.z  -  z * w.y;
    REAL8m ty = z * w.x  -  x * w.z;
    REAL8m tz = x * w.y  -  y * w.x;
    x = tx; y = ty; z = tz;
#else // In principle, the ordering below reduces register pressure
    REAL8m p12 = x*w.y;
    REAL8m p13 = x*w.z;
    REAL8m p23 = y*w.z;
    p12 -= y*w.x;
    y = z*w.x - p13;
    x = p23 - z*w.y;
    z = p12;
#endif

    return *this;
  }

  Oxs_ThreeVector& operator*=(REAL8m a) { x*=a; y*=a; z*=a; return *this; }

  Oxs_ThreeVector Accum(REAL8m a,const Oxs_ThreeVector& v) {
    x+=a*v.x;  y+=a*v.y;  z+=a*v.z;  return *this;
  }


  Oxs_ThreeVector& wxvxw(const Oxs_ThreeVector& w) {
    // Performs w x *this x w

    REAL8m tx = y * w.z  -  z * w.y;
    REAL8m ty = z * w.x  -  x * w.z;
    z = ty * w.x  - tx * w.y ;

    REAL8m tz = x * w.y  -  y * w.x;
    x = tz * w.y  - ty * w.z ;
    y = tx * w.z  - tz * w.x ;

    return *this;
  }

  // Misc
  REAL8m MagSq() const { return x*x+y*y+z*z; }
  REAL8m TaxicabNorm() const { return fabs(x)+fabs(y)+fabs(z); }  // aka L1-norm
  void SetMag(REAL8m mag);  // Adjusts size to "mag"
  void Random(REAL8m mag);  // Makes a random vector of size "mag"
  REAL8m MakeUnit(); // High-precision version of SetMag(1.0).
  /// Return value is original MagSq(), i.e., on entry.
};

// Test operators on Oxs_ThreeVector's
inline BOOL operator==(const Oxs_ThreeVector& lhs,
		const Oxs_ThreeVector& rhs)
{ return ((lhs.x==rhs.x) && (lhs.y==rhs.y) && (lhs.z==rhs.z)); }

inline BOOL operator!=(const Oxs_ThreeVector& lhs,
		const Oxs_ThreeVector& rhs)
{ return ((lhs.x!=rhs.x) || (lhs.y!=rhs.y) || (lhs.z!=rhs.z)); }

// The next two operators are defined so MSVC++ 5.0 will accept
// vector<ThreeVector>, but are left undefined because there
// is no way to define them that makes sense in all contexts.
BOOL operator<(const Oxs_ThreeVector&,const Oxs_ThreeVector&);
BOOL operator>(const Oxs_ThreeVector&,const Oxs_ThreeVector&);


// Binary operators on Oxs_ThreeVector's
inline const Oxs_ThreeVector
operator+(const Oxs_ThreeVector& lhs,
	  const Oxs_ThreeVector& rhs)
{ Oxs_ThreeVector result(lhs); return result+=rhs; }

inline const Oxs_ThreeVector
operator-(const Oxs_ThreeVector& lhs,
	  const Oxs_ThreeVector& rhs)
{ Oxs_ThreeVector result(lhs); return result-=rhs; }

// Cross product
inline const Oxs_ThreeVector
operator^(const Oxs_ThreeVector& v,
	  const Oxs_ThreeVector& w)
{
#if 0
  REAL8m tx = v.y * w.z  -  w.y * v.z;
  REAL8m ty = w.x * v.z  -  v.x * w.z;
  REAL8m tz = v.x * w.y  -  w.x * v.y;
#else // In principle, the ordering below reduces register pressure
  REAL8m tz = v.x*w.y;
  REAL8m ty = v.x*w.z;
  REAL8m tx = v.y*w.z;
  tz -= v.y*w.x;
  ty  = v.z*w.x - ty;
  tx -= v.z*w.y;
#endif
  return Oxs_ThreeVector(tx,ty,tz);
}


// Product against scalar
inline const Oxs_ThreeVector
operator*(REAL8m scalar,
	  const Oxs_ThreeVector& vec)
{ Oxs_ThreeVector result(vec); return result*=scalar; }

inline const Oxs_ThreeVector
operator*(const Oxs_ThreeVector& vec,
	  REAL8m scalar)
{ Oxs_ThreeVector result(vec); return result*=scalar; }


// Dot product
inline REAL8m
operator*(const Oxs_ThreeVector& lhs,const Oxs_ThreeVector& rhs)
{ return lhs.x*rhs.x + lhs.y*rhs.y + lhs.z*rhs.z; }

#endif // _OXS_THREEVECTOR
