/* FILE: exchange6ngbr.cc            -*-Mode: c++-*-
 *
 * 6 neighbor exchange energy on rectangular mesh,
 * derived from Oxs_Energy class.
 *
 */

#include <string>

#include "atlas.h"
#include "nb.h"
#include "key.h"
#include "director.h"
#include "mesh.h"
#include "meshvalue.h"
#include "oxswarn.h"
#include "simstate.h"
#include "threevector.h"
#include "rectangularmesh.h"
#include "exchange6ngbr.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

OC_USE_STRING;

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_Exchange6Ngbr);

/* End includes */

// Revision information, set via CVS keyword substitution
static const Oxs_WarningMessageRevisionInfo revision_info
  (__FILE__,
   "$Revision: 1.37 $",
   "$Date: 2009-11-23 07:42:41 $",
   "$Author: donahue $",
   "Michael J. Donahue (michael.donahue@nist.gov)");

// Constructor
Oxs_Exchange6Ngbr::Oxs_Exchange6Ngbr(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_ChunkEnergy(name,newdtr,argstr),
    coef_size(0), coef(NULL), mesh_id(0)
{
  // Process arguments
  OXS_GET_INIT_EXT_OBJECT("atlas",Oxs_Atlas,atlas);
  atlaskey.Set(atlas.GetPtr());
  /// Dependency lock is held until *this is deleted.

  // Determine number of regions, and check that the
  // count lies within the allowed range.
  coef_size = atlas->GetRegionCount();
  if(coef_size<1) {
    String msg = String("Oxs_Atlas object ")
      + atlas->InstanceName()
      + String(" must contain at least one region.");

    throw Oxs_ExtError(msg.c_str());
  }
  UINT4m i;
  const UINT4m sqrootbits = (sizeof(UINT4m)*8)/2;
  UINT4m coef_check = 1;
  for(i=0;i<sqrootbits;++i) coef_check *= 2;
  const UINT4m max_coef_size = coef_check - 1;
  if(coef_size>max_coef_size) {
      char buf[4096];
      Oc_Snprintf(buf,sizeof(buf),
		  "Oxs_Atlas object %s has too many regions: %lu"
                  " (max allowed: %lu)",
                  atlas->InstanceName(),
                  (unsigned long)coef_size,
                  (unsigned long)max_coef_size);
      throw Oxs_ExtError(this,buf);
  }

  // Determine coef matrix fill type
  String typestr;
  vector<String> params;
  REAL8m default_coef = 0.0;
  BOOL has_A = HasInitValue("A");
  BOOL has_lex = HasInitValue("lex");
  if(has_A && has_lex) {
    throw Oxs_ExtError(this,"Invalid exchange coefficient request:"
			 " both A and lex specified; only one should"
			 " be given.");
  } else if(has_lex) {
    excoeftype = LEX_TYPE;
    typestr = "lex";
    default_coef = GetRealInitValue("default_lex",0.0);
    FindRequiredInitValue("lex",params);
  } else {
    excoeftype = A_TYPE;
    typestr = "A";
    default_coef = GetRealInitValue("default_A",0.0);
    FindRequiredInitValue("A",params);
  }
  if(params.empty()) {
    String msg = String("Empty parameter list for key \"")
      + typestr + String("\"");
    throw Oxs_ExtError(this,msg);
  }
  if(params.size()%3!=0) {
      char buf[4096];
      Oc_Snprintf(buf,sizeof(buf),
		  "Number of elements in %s sub-list must be"
		  " divisible by 3"
		  " (actual sub-list size: %u)",
		  typestr.c_str(),(unsigned int)params.size());
      throw Oxs_ExtError(this,buf);
  }

  // Allocate A matrix.  Because raw pointers are used, a memory leak
  // would occur if an uncaught exception occurred beyond this point.
  // So, we put the whole bit in a try-catch block.
  try {
    coef = new REAL8m*[coef_size];
    coef[0] = NULL; // Safety, in case next alloc throws an exception
    coef[0] = new REAL8m[coef_size*coef_size];
    for(i=1;i<coef_size;i++) coef[i] = coef[i-1] + coef_size;

    // Fill A matrix
    for(i=0;i<coef_size*coef_size;i++) coef[0][i] = default_coef;
    for(i=0;i<params.size();i+=3) {
      INT4m i1 = atlas->GetRegionId(params[i]);
      INT4m i2 = atlas->GetRegionId(params[i+1]);
      if(i1<0 || i2<0) {
        // Unknown region(s) requested
        char buf[4096];
        char* cptr=buf;
        if(i1<0) {
          Oc_Snprintf(buf,sizeof(buf),
                      "First entry in %s[%u] sub-list, \"%s\","
                      " is not a known region in atlas \"%s\".  ",
                      typestr.c_str(),i/3,params[i].c_str(),
                      atlas->InstanceName());
          cptr += strlen(buf);
        }
        if(i2<0) {
          Oc_Snprintf(cptr,sizeof(buf)-(cptr-buf),
                      "Second entry in %s[%u] sub-list, \"%s\","
                      " is not a known region in atlas \"%s\".  ",
                      typestr.c_str(),i/3,params[i+1].c_str(),
                      atlas->InstanceName());
        }
        String msg = String(buf);
        msg += String("Known regions:");
        vector<String> regions;
        atlas->GetRegionList(regions);
        for(unsigned int j=0;j<regions.size();++j) {
          msg += String(" \n");
          msg += regions[j];
        }
        throw Oxs_ExtError(this,msg);
      }
      BOOL err;
      REAL8m coefpair = Nb_Atof(params[i+2].c_str(),err);
      if(err) {
        char buf[4096];
        Oc_Snprintf(buf,sizeof(buf),
                    "Third entry in %s[%u] sub-list, \"%s\","
                    " is not a valid floating point number.",
                    typestr.c_str(),i/3,params[i+2].c_str());
        throw Oxs_ExtError(this,buf);
      }
      coef[i1][i2]=coefpair;
      coef[i2][i1]=coefpair; // coef should be symmetric
    }
    DeleteInitValue(typestr);

    VerifyAllInitArgsUsed();

    // Setup outputs
    maxspinangle_output.Setup(this,InstanceName(),"Max Spin Ang","deg",1,
                              &Oxs_Exchange6Ngbr::UpdateDerivedOutputs);
    maxspinangle_output.Register(director,0);
    stage_maxspinangle_output.Setup(this,InstanceName(),"Stage Max Spin Ang","deg",1,
                                    &Oxs_Exchange6Ngbr::UpdateDerivedOutputs);
    stage_maxspinangle_output.Register(director,0);
    run_maxspinangle_output.Setup(this,InstanceName(),"Run Max Spin Ang","deg",1,
                                  &Oxs_Exchange6Ngbr::UpdateDerivedOutputs);
    run_maxspinangle_output.Register(director,0);
  }
  catch(...) {
    if(coef_size>0 && coef!=NULL) { // Release coef memory
      delete[] coef[0]; delete[] coef;
      coef = NULL;
      coef_size = 0; // Safety
    }
    throw;
  }

}

Oxs_Exchange6Ngbr::~Oxs_Exchange6Ngbr()
{
  if(coef_size>0 && coef!=NULL) {
    delete[] coef[0];
    delete[] coef;
  }
}

BOOL Oxs_Exchange6Ngbr::Init()
{
  mesh_id = 0;
  region_id.Release();
  return Oxs_Energy::Init();
}

void Oxs_Exchange6Ngbr::CalcEnergyA
(const Oxs_SimState& state,
 Oxs_ComputeEnergyDataThreaded& ocedt,
 Oxs_ComputeEnergyDataThreadedAux& ocedtaux,
 UINT4m node_start,
 UINT4m node_stop,
 int threadnumber
 ) const
{
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);

  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(state.mesh);
  if(mesh==NULL) {
    String msg = String("Import mesh to Oxs_Exchange6Ngbr::GetEnergyA()"
                        " routine of object ")
      + String(InstanceName())
      + String(" is not an Oxs_RectangularMesh object.");
    throw Oxs_ExtError(msg.c_str());
  }

  UINT4m xdim = mesh->DimX();
  UINT4m ydim = mesh->DimY();
  UINT4m zdim = mesh->DimZ();
  UINT4m xydim = xdim*ydim;

  REAL8m wgtx = -1.0/(mesh->EdgeLengthX()*mesh->EdgeLengthX());
  REAL8m wgty = -1.0/(mesh->EdgeLengthY()*mesh->EdgeLengthY());
  REAL8m wgtz = -1.0/(mesh->EdgeLengthZ()*mesh->EdgeLengthZ());

  REAL8m hcoef = -2/MU0;

  Nb_Xpfloat energy_sum = 0;
  REAL8m thread_mindot = mindot[threadnumber];
  // Note: For mindot (equivalently, maxangle) calculation, it suffices
  // to check spin[j]*spin[i] for j>i.

  UINT4m x,y,z;
  mesh->GetCoords(node_start,x,y,z);

  UINT4m i = node_start;
  while(i<node_stop) {
    UINT4m xstop = xdim;
    if(xdim-x>node_stop-i) xstop = x + (node_stop-i);
    while(x<xstop) {
      ThreeVector base = spin[i];
      REAL8m Msii = Ms_inverse[i];
      if(0.0 == Msii) {
        if(ocedt.energy) (*ocedt.energy)[i] = 0.0;
        if(ocedt.H)      (*ocedt.H)[i].Set(0.,0.,0.);
        ++i;   ++x;
        continue;
      }
      REAL8m* Arow = coef[region_id[i]];
      ThreeVector sum(0.,0.,0.);
      if(z>0) {
        UINT4m j = i-xydim;
        REAL8m Apair = Arow[region_id[j]];
        if(Apair!=0 && Ms_inverse[j]!=0.0) {
          sum += Apair*wgtz*(spin[j] - base);
          REAL8m dot = spin[j]*base;
          if(dot<thread_mindot) thread_mindot = dot;
        }
      }
      if(y>0) {
        UINT4m j = i-xdim;
        REAL8m Apair = Arow[region_id[j]];
        if(Apair!=0.0 && Ms_inverse[j]!=0.0) {
          sum += Apair*wgty*(spin[j] - base);
          REAL8m dot = spin[j]*base;
          if(dot<thread_mindot) thread_mindot = dot;
        }
      }
      if(x>0) {
        UINT4m j = i-1;
        REAL8m Apair = Arow[region_id[j]];
        if(Apair!=0.0 && Ms_inverse[j]!=0.0) {
          sum += Apair*wgtx*(spin[j] - base);
          REAL8m dot = spin[j]*base;
          if(dot<thread_mindot) thread_mindot = dot;
        }
      }
      if(x<xdim-1) {
        UINT4m j = i+1;
        REAL8m Apair = Arow[region_id[j]];
        if(Ms_inverse[j]!=0.0) sum += Apair*wgtx*(spin[j] - base);
      }
      if(y<ydim-1) {
        UINT4m j = i+xdim;
        REAL8m Apair = Arow[region_id[j]];
        if(Ms_inverse[j]!=0.0) sum += Apair*wgty*(spin[j] - base);
      }
      if(z<zdim-1) {
        UINT4m j = i+xydim;
        REAL8m Apair = Arow[region_id[j]];
        if(Ms_inverse[j]!=0.0) sum += Apair*wgtz*(spin[j]- base);
      }

      REAL8m ei = base.x*sum.x + base.y*sum.y + base.z*sum.z;
      REAL8m hmult = hcoef*Msii;
      sum.x *= hmult;  sum.y *= hmult;   sum.z *= hmult;
      REAL8m tx = base.y*sum.z - base.z*sum.y;
      REAL8m ty = base.z*sum.x - base.x*sum.z;
      REAL8m tz = base.x*sum.y - base.y*sum.x;

      energy_sum += ei;
      if(ocedt.energy)       (*ocedt.energy)[i] = ei;
      if(ocedt.energy_accum) (*ocedt.energy_accum)[i] += ei;
      if(ocedt.H)       (*ocedt.H)[i] = sum;
      if(ocedt.H_accum) (*ocedt.H_accum)[i] += sum;
      if(ocedt.mxH)       (*ocedt.mxH)[i] = ThreeVector(tx,ty,tz);
      if(ocedt.mxH_accum) (*ocedt.mxH_accum)[i] += ThreeVector(tx,ty,tz);
      ++i;   ++x;
    }
    x=0;
    if((++y)>=ydim) {
      y=0;
      ++z;
    }
  }

  ocedtaux.energy_total_accum += energy_sum.GetValue() * mesh->Volume(0);
  /// All cells have same volume in an Oxs_RectangularMesh.

  mindot[threadnumber] = thread_mindot;
}


void Oxs_Exchange6Ngbr::CalcEnergyLex
(const Oxs_SimState& state,
 Oxs_ComputeEnergyDataThreaded& ocedt,
 Oxs_ComputeEnergyDataThreadedAux& ocedtaux,
 UINT4m node_start,
 UINT4m node_stop,
 int threadnumber
 ) const
{
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);

  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(state.mesh);
  if(mesh==NULL) {
    String msg = String("Import mesh to Oxs_Exchange6Ngbr::GetEnergyLex()"
                        " routine of object ")
      + String(InstanceName())
      + String(" is not an Oxs_RectangularMesh object.");
    throw Oxs_ExtError(msg);
  }

  UINT4m xdim = mesh->DimX();
  UINT4m ydim = mesh->DimY();
  UINT4m zdim = mesh->DimZ();
  UINT4m xydim = xdim*ydim;

  REAL8m wgtx = 1.0/(mesh->EdgeLengthX()*mesh->EdgeLengthX());
  REAL8m wgty = 1.0/(mesh->EdgeLengthY()*mesh->EdgeLengthY());
  REAL8m wgtz = 1.0/(mesh->EdgeLengthZ()*mesh->EdgeLengthZ());

  Nb_Xpfloat energy_sum = 0;
  REAL8m thread_mindot = mindot[threadnumber];
  // Note: For mindot (equivalently, maxangle) calculation, it suffices
  // to check spin[j]*spin[i] for j>i, or j<i, or various mixes of the
  // two.

  UINT4m x,y,z;
  mesh->GetCoords(node_start,x,y,z);

  UINT4m i = node_start;
  while(i<node_stop) {
    UINT4m xstop = xdim;
    if(xdim-x>node_stop-i) xstop = x + (node_stop-i);
    while(x<xstop) {
      if(0 == Ms[i]) {
        if(ocedt.energy) (*ocedt.energy)[i] = 0.0;
        if(ocedt.H)      (*ocedt.H)[i].Set(0.,0.,0.);
        ++i;   ++x;
        continue;
      }
      const REAL8m emult = (-0.5*MU0)*Ms[i];

      ThreeVector base = spin[i];

      REAL8m* lexrow = coef[region_id[i]];
      ThreeVector sum(0.,0.,0.);
      if(z>0) {
        UINT4m j = i-xydim;
        REAL8m lexpair = lexrow[region_id[j]];
        if(lexpair!=0.0 && Ms[j]!=0.0) {
          sum += Ms[j]*lexpair*lexpair*wgtz*(spin[j] - base);
          REAL8m dot = spin[j]*base;
          if(dot<thread_mindot) thread_mindot = dot;
        }
      }
      if(y>0) {
        UINT4m j = i-xdim;
        REAL8m lexpair = lexrow[region_id[j]];
        if(lexpair!=0.0 && Ms[j]!=0.0) {
          sum += Ms[j]*lexpair*lexpair*wgty*(spin[j] - base);
          REAL8m dot = spin[j]*base;
          if(dot<thread_mindot) thread_mindot = dot;
        }
      }
      if(x>0) {
        UINT4m j = i-1;
        REAL8m lexpair = lexrow[region_id[j]];
        if(lexpair!=0.0 && Ms[j]!=0.0) {
          sum += Ms[j]*lexpair*lexpair*wgtx*(spin[j] - base);
          REAL8m dot = spin[j]*base;
          if(dot<thread_mindot) thread_mindot = dot;
        }
      }
      if(x<xdim-1) {
        UINT4m j = i+1;
        REAL8m lexpair = lexrow[region_id[j]];
        sum += Ms[j]*lexpair*lexpair*wgtx*(spin[j] - base);
      }
      if(y<ydim-1) {
        UINT4m j = i+xdim;
        REAL8m lexpair = lexrow[region_id[j]];
        sum += Ms[j]*lexpair*lexpair*wgty*(spin[j] - base);
      }
      if(z<zdim-1) {
        UINT4m j = i+xydim;
        REAL8m lexpair = lexrow[region_id[j]];
        sum += Ms[j]*lexpair*lexpair*wgtz*(spin[j]- base);
      }

      REAL8m ei = emult*(base.x*sum.x + base.y*sum.y + base.z*sum.z);
      REAL8m tx = base.y*sum.z - base.z*sum.y;
      REAL8m ty = base.z*sum.x - base.x*sum.z;
      REAL8m tz = base.x*sum.y - base.y*sum.x;

      energy_sum += ei;
      if(ocedt.energy)       (*ocedt.energy)[i] = ei;
      if(ocedt.energy_accum) (*ocedt.energy_accum)[i] += ei;
      if(ocedt.H)       (*ocedt.H)[i] = sum;
      if(ocedt.H_accum) (*ocedt.H_accum)[i] += sum;
      if(ocedt.mxH)       (*ocedt.mxH)[i] = ThreeVector(tx,ty,tz);
      if(ocedt.mxH_accum) (*ocedt.mxH_accum)[i] += ThreeVector(tx,ty,tz);
      ++i;   ++x;
    }
    x=0;
    if((++y)>=ydim) {
      y=0;
      ++z;
    }
  }

  ocedtaux.energy_total_accum += energy_sum.GetValue() * mesh->Volume(0);
  /// All cells have same volume in an Oxs_RectangularMesh.

  mindot[threadnumber] = thread_mindot;
}


void Oxs_Exchange6Ngbr::ComputeEnergyChunkInitialize
(const Oxs_SimState& /* state */,
 Oxs_ComputeEnergyDataThreaded& /* ocedt */,
 Oxs_ComputeEnergyDataThreadedAux& /* ocedtaux */,
 int number_of_threads) const
{
  if(mindot.size() != (vector<REAL8m>::size_type)number_of_threads) {
    mindot.resize(number_of_threads);
  }
  for(int i=0;i<number_of_threads;++i) {
    mindot[i] = 1.0; // Maximum possible value for m_i*m_j
  }
}


void Oxs_Exchange6Ngbr::ComputeEnergyChunkFinalize
(const Oxs_SimState& state,
 const Oxs_ComputeEnergyDataThreaded& /* ocedt */,
 const Oxs_ComputeEnergyDataThreadedAux& /* ocedtaux */,
 int number_of_threads) const
{
  // Set max angle data
  REAL8m total_mindot = 1.0;
  for(int i=0;i<number_of_threads;++i) {
    if(mindot[i]<total_mindot) total_mindot = mindot[i];
  }
  REAL8m maxang;
    // Handle extremal cases separately in case spins drift off S2
  if(total_mindot >  1 - 8*REAL8_EPSILON)      maxang = 0.0;
  else if(total_mindot < -1 + 8*REAL8_EPSILON) maxang = 180.0;
  else                  maxang = acos(total_mindot)*(180.0/PI);

  REAL8m dummy_value;
  String msa_name = MaxSpinAngleStateName();
  if(state.GetDerivedData(msa_name,dummy_value)) {
    // Ideally, energy values would never be computed more than once
    // for any one state, but in practice it seems inevitable that
    // such will occur on occasion.  For example, suppose a user
    // requests output on a state obtained by a stage crossing (as
    // opposed to a state obtained through a normal intrastage step);
    // a subsequent ::Step operation will re-compute the energies
    // because not all the information needed by the step transition
    // machinery is cached from an energy computation.  Even user
    // output requests on post-::Step states is problematic if some of
    // the requested output is not cached as part of the step
    // proceedings.  A warning is put into place below for debugging
    // purposes, but in general an error is raised only if results
    // from the recomputation are different than originally.
#ifndef NDEBUG
    static Oxs_WarningMessage maxangleset(3);
    maxangleset.Send(revision_info,OC_STRINGIFY(__LINE__),
                     "Programming error?"
                     " Oxs_Exchange6Ngbr max spin angle set twice.");
#endif
    if(fabs(dummy_value-maxang)>8*180.*REAL8_EPSILON) {
      char errbuf[1024];
      Oc_Snprintf(errbuf,sizeof(errbuf),
                  "Programming error:"
                  " Oxs_Exchange6Ngbr max spin angle set to"
                  " two different values;"
                  " orig val=%#.17g, new val=%#.17g",
                  dummy_value,maxang);
      throw Oxs_ExtError(this,errbuf);
    }
  } else {
    state.AddDerivedData(msa_name,maxang);
  }

  // Run and stage angle data depend on data from the previous state.
  // In the case that the energy (and hence max stage and run angle)
  // for the current state was computed previously, then the previous
  // state may have been dropped.  So, compute and save run and stage
  // angle data iff they are not already computed.

  // Check stage and run max angle data from previous state
  const Oxs_SimState* oldstate = NULL;
  REAL8m stage_maxang = -1;
  REAL8m run_maxang = -1;
  String smsa_name = StageMaxSpinAngleStateName();
  String rmsa_name = RunMaxSpinAngleStateName();
  if(state.previous_state_id &&
     (oldstate
      = director->FindExistingSimulationState(state.previous_state_id)) ) {
    if(oldstate->stage_number != state.stage_number) {
      stage_maxang = 0.0;
    } else {
      if(oldstate->GetDerivedData(smsa_name,dummy_value)) {
        stage_maxang = dummy_value;
      }
    }
    if(oldstate->GetDerivedData(rmsa_name,dummy_value)) {
      run_maxang = dummy_value;
    }
  }
  if(stage_maxang<maxang) stage_maxang = maxang;
  if(run_maxang<maxang)   run_maxang = maxang;

  // Stage max angle data
  if(!state.GetDerivedData(smsa_name,dummy_value)) {
    state.AddDerivedData(smsa_name,stage_maxang);
  }

  // Run max angle data
  if(!state.GetDerivedData(rmsa_name,dummy_value)) {
    state.AddDerivedData(rmsa_name,run_maxang);
  }
}


void Oxs_Exchange6Ngbr::ComputeEnergyChunk
(const Oxs_SimState& state,
 Oxs_ComputeEnergyDataThreaded& ocedt,
 Oxs_ComputeEnergyDataThreadedAux& ocedtaux,
 UINT4m node_start,
 UINT4m node_stop,
 int threadnumber
 ) const
{
#ifndef NDEBUG
  if(node_stop>state.mesh->Size() || node_start>node_stop) {
    throw Oxs_ExtError(this,"Programming error:"
                       " Invalid node_start/node_stop values");
  }
#endif

  const UINT4m size = state.mesh->Size();
  if(size<1) {
    return;
  }

  if(mesh_id !=  state.mesh->Id() || !atlaskey.SameState()) {
    // Setup region mapping.
    // NB: At a lower level, this may potentially involve calls back
    // into the Tcl interpreter.  Per Tcl spec, only the thread
    // originating the interpreter is allowed to make calls into it, so
    // only threadnumber == 0 can do this processing.  Any other thread
    // must block until that processing is complete.
    thread_control.Lock();
    if(Oxs_ThreadError::IsError()) {
      if(thread_control.count>0) {
        // Release a blocked thread
        thread_control.Notify();
      }
      thread_control.Unlock();
      return; // What else?
    }
    if(threadnumber != 0) {
      if(mesh_id != state.mesh->Id() || !atlaskey.SameState()) {
        // If above condition is false, then the main thread came
        // though and initialized everything between the time of
        // the previous check and this thread's acquiring of the
        // thread_control mutex; in which case, "never mind".
        // Otherwise:
        ++thread_control.count; // Multiple threads may progress to this
        /// point before the main thread (threadnumber == 0) grabs the
        /// thread_control mutex.  Keep track of how many, so that
        /// afterward they may be released, one by one.  (The main
        /// thread will Notify control_wait.cond once; after that
        /// as each waiting thread is released, the newly released
        /// thread sends a Notify to wake up the next one.
        thread_control.Wait(0);
        --thread_control.count;
        int condcheckerror=0;
        if(mesh_id !=  state.mesh->Id() || !atlaskey.SameState()) {
          // Error?
          condcheckerror=1;
          Oxs_ThreadPrintf(stderr,"Invalid condition in"
                           " Oxs_Exchange6Ngbr::ComputeEnergyChunk(),"
                           " thread number %d\n",threadnumber);
        }
        if(thread_control.count>0) {
          // Free a waiting thread.
          thread_control.Notify();
        }
        thread_control.Unlock();
        if(condcheckerror || Oxs_ThreadError::IsError()) {
          return; // What else?
        }
      } else {
        if(thread_control.count>0) {
          // Free a waiting thread.  (Actually, it can occur that the
          // thread_control will be grabbed by another thread that is
          // blocked at the first thread_control mutex Lock() call above
          // rather than on the ConditionWait, in which case this
          // ConditionNotify will be effectively lost.  But that is
          // okay, because then *that* thread will Notify when it
          // releases the mutex.)
          thread_control.Notify();
        }
        thread_control.Unlock();
      }
    } else {
      // Main thread (threadnumber == 0)
      try {
        region_id.AdjustSize(state.mesh);
        ThreeVector location;
        for(UINT4m i=0;i<size;i++) {
          state.mesh->Center(i,location);
          if((region_id[i] = atlas->GetRegionId(location))<0) {
            String msg = String("Import mesh to Oxs_Exchange6Ngbr::GetEnergy()"
                                " routine of object ")
              + String(InstanceName())
                + String(" has points outside atlas ")
              + String(atlas->InstanceName());
            throw Oxs_ExtError(msg.c_str());
          }
        }
        atlaskey.Set(atlas.GetPtr());
        mesh_id = state.mesh->Id();
      } catch(Oxs_ExtError& err) {
        // Leave unmatched mesh_id as a flag to check
        // Oxs_ThreadError for an error.
        Oxs_ThreadError::SetError(String(err));
        if(thread_control.count>0) {
          thread_control.Notify();
        }
        thread_control.Unlock();
        throw;
      } catch(String& serr) {
        // Leave unmatched mesh_id as a flag to check
        // Oxs_ThreadError for an error.
        Oxs_ThreadError::SetError(serr);
        if(thread_control.count>0) {
          thread_control.Notify();
        }
        thread_control.Unlock();
        throw;
      } catch(const char* cerr) {
        // Leave unmatched mesh_id as a flag to check
        // Oxs_ThreadError for an error.
        Oxs_ThreadError::SetError(String(cerr));
        if(thread_control.count>0) {
          thread_control.Notify();
        }
        thread_control.Unlock();
        throw;
      } catch(...) {
        // Leave unmatched mesh_id as a flag to check
        // Oxs_ThreadError for an error.
        Oxs_ThreadError::SetError(String("Error in "
            "Oxs_Exchange6Ngbr::ComputeEnergyChunk"));
        if(thread_control.count>0) {
          thread_control.Notify();
        }
        thread_control.Unlock();
        throw;
      }
      if(thread_control.count>0) {
        // Free a waiting thread.  (Actually, it can occur that the
        // thread_control will be grabbed by another thread that is
        // blocked at the first thread_control mutex Lock() call above
        // rather than on the ConditionWait, in which case this
        // ConditionNotify will be effectively lost.  But that is
        // okay, because then *that* thread will Notify when it
        // releases the mutex.)
        thread_control.Notify();
      }
      thread_control.Unlock();
    }
  }
  if(excoeftype == LEX_TYPE) {
    CalcEnergyLex(state,ocedt,ocedtaux,node_start,node_stop,threadnumber);
  } else {
    CalcEnergyA(state,ocedt,ocedtaux,node_start,node_stop,threadnumber);
  }
}

void Oxs_Exchange6Ngbr::UpdateDerivedOutputs(const Oxs_SimState& state)
{
  maxspinangle_output.cache.state_id
    = stage_maxspinangle_output.cache.state_id
    = run_maxspinangle_output.cache.state_id
    = 0;  // Mark change in progress

  String dummy_name = MaxSpinAngleStateName();
  if(!state.GetDerivedData(dummy_name.c_str(),
                           maxspinangle_output.cache.value)) {
    // Error; This should always be set.  For now, just set the value to
    // -1, but in the future should consider throwing an exception.
    maxspinangle_output.cache.value = -1.0;
  }

  String stage_dummy_name = StageMaxSpinAngleStateName();
  if(!state.GetDerivedData(stage_dummy_name.c_str(),
                           stage_maxspinangle_output.cache.value)) {
    // Error; This should always be set.  For now, just set the value to
    // -1, but in the future should consider throwing an exception.
    stage_maxspinangle_output.cache.value = -1.0;
  }

  String run_dummy_name = RunMaxSpinAngleStateName();
  if(!state.GetDerivedData(run_dummy_name.c_str(),
                           run_maxspinangle_output.cache.value)) {
    // Error; This should always be set.  For now, just set the value to
    // -1, but in the future should consider throwing an exception.
    run_maxspinangle_output.cache.value = -1.0;
  }

  maxspinangle_output.cache.state_id
    = stage_maxspinangle_output.cache.state_id
    = run_maxspinangle_output.cache.state_id
    = state.Id();
}
